import os
import pytest

from sismic.io import import_from_yaml
from sismic.interpreter import Interpreter



@pytest.fixture(params=[False, True], ids=['no contract', 'contract'])
def elevator(request):
    if request.param:
        sc = import_from_yaml(filepath='docs/examples/elevator/elevator_contract.yaml')
    else:
        sc = import_from_yaml(filepath='docs/examples/elevator/elevator.yaml')

    return Interpreter(sc)


@pytest.fixture
def remote_elevator(elevator):
    sc = import_from_yaml(filepath='docs/examples/elevator/elevator_buttons.yaml')
    remote = Interpreter(sc)
    remote.bind(elevator)
    return remote


@pytest.fixture
def writer():
    sc = import_from_yaml(filepath='docs/examples/writer_options.yaml')
    return Interpreter(sc)


@pytest.fixture(params=[False, True], ids=['no contract', 'contract'])
def microwave(request):
    if request.param:
        sc = import_from_yaml(filepath='docs/examples/microwave/microwave_with_contracts.yaml')
    else:
        sc = import_from_yaml(filepath='docs/examples/microwave/microwave.yaml')
    return Interpreter(sc)


@pytest.fixture
def simple_statechart():
    return import_from_yaml(filepath='tests/yaml/simple.yaml')


@pytest.fixture
def composite_statechart():
    return import_from_yaml(filepath='tests/yaml/composite.yaml')


@pytest.fixture
def deep_history_statechart():
    return import_from_yaml(filepath='tests/yaml/deep_history.yaml')


@pytest.fixture
def final_statechart():
    return import_from_yaml(filepath='tests/yaml/final.yaml')


@pytest.fixture
def infinite_statechart():
    return import_from_yaml(filepath='tests/yaml/infinite.yaml')


@pytest.fixture
def parallel_statechart():
    return import_from_yaml(filepath='tests/yaml/parallel.yaml')


@pytest.fixture
def nested_parallel_statechart():
    return import_from_yaml(filepath='tests/yaml/nested_parallel.yaml')


@pytest.fixture
def nondeterministic_statechart():
    return import_from_yaml(filepath='tests/yaml/nondeterministic.yaml')


@pytest.fixture
def history_statechart():
    return import_from_yaml(filepath='tests/yaml/history.yaml')


@pytest.fixture
def internal_statechart():
    return import_from_yaml(filepath='tests/yaml/internal.yaml')


@pytest.fixture
def priority_statechart():
    return import_from_yaml(filepath='tests/yaml/priority.yaml')


@pytest.fixture(params=['actions', 'composite', 'history', 'deep_history', 'final', 'infinite', 'internal', 'priority',
                        'nested_parallel', 'nondeterministic', 'parallel', 'simple', 'timer'])
def example_from_tests(request):
    return import_from_yaml(filepath=os.path.join('tests', 'yaml', request.param + '.yaml'))


@pytest.fixture(params=['elevator/elevator', 'elevator/elevator_contract', 'microwave/microwave',
                        'elevator/tester_elevator_7th_floor_never_reached', 'elevator/tester_elevator_moves_after_10s',
                        'writer_options'])
def example_from_docs(request):
    return import_from_yaml(filepath=os.path.join('docs', 'examples', request.param + '.yaml'))



from typing import Any, List, Mapping, MutableMapping, Optional, Tuple, cast

from ..exceptions import StatechartError
from ..model import (ActionStateMixin, BasicState, CompositeStateMixin,
                     CompoundState, DeepHistoryState, FinalState,
                     OrthogonalState, ShallowHistoryState, Statechart,
                     StateMixin, Transition, TransitionStateMixin)

__all__ = ['import_from_dict', 'export_to_dict']


def import_from_dict(data: Mapping[str, Any]) -> Statechart:
    data = data['statechart']

    statechart = Statechart(name=data['name'],
                            description=data.get('description', None),
                            preamble=data.get('preamble', None))

    states = []  # (StateMixin instance, parent name)
    transitions = []  # Transition instances
    # (State dict, parent name)
    # type: List[Tuple[Mapping[str, Any], Optional[str]]]
    data_to_consider = [(data['root state'], None)]

    while data_to_consider:
        state_data, parent_name = data_to_consider.pop()

        # Get state
        try:
            state = _import_state_from_dict(state_data)
        except StatechartError:
            raise
        except Exception as e:
            raise StatechartError('Unable to load given YAML') from e
        states.append((state, parent_name))

        # Get substates
        if isinstance(state, CompoundState):
            for substate_data in state_data['states']:
                data_to_consider.append((substate_data, state.name))
        elif isinstance(state, OrthogonalState):
            for substate_data in state_data['parallel states']:
                data_to_consider.append((substate_data, state.name))

        # Get transition(s)
        for transition_data in state_data.get('transitions', []):
            try:
                transition = _import_transition_from_dict(state.name, transition_data)
            except StatechartError:
                raise
            except Exception as e:
                raise StatechartError('Unable to load given YAML') from e
            transitions.append(transition)

    # Register on statechart
    for state, parent in states:
        statechart.add_state(state, parent)
    for transition in transitions:
        statechart.add_transition(transition)

    return statechart


def _import_transition_from_dict(state_name: str, transition_d: Mapping[str, Any]) -> Transition:
    """
    Return a Transition instance from given dict.

    :param state_name: name of the state in which the transition is defined
    :param transition_d: a dictionary containing transition data
    :return: an instance of Transition
    """
    event = transition_d.get('event', None)
    guard = transition_d.get('guard', None)
    action = transition_d.get('action', None)
    priority = transition_d.get('priority', None)

    if priority == 'low':
        priority = Transition.LOW_PRIORITY
    elif priority == 'high':
        priority = Transition.HIGH_PRIORITY

    transition = Transition(
        state_name,
        transition_d.get('target', None),
        event.strip() if event else None,
        guard.strip() if guard else None,
        action.strip() if action else None,
        priority,
    )

    # Preconditions, postconditions and invariants
    for condition in transition_d.get('contract', []):
        if condition.get('before', None):
            transition.preconditions.append(condition['before'].strip())
        elif condition.get('after', None):
            transition.postconditions.append(condition['after'].strip())
        elif condition.get('always', None):
            transition.invariants.append(condition['always'].strip())

    return transition


def _import_state_from_dict(state_d: Mapping[str, Any]) -> StateMixin:
    """
    Return the appropriate type of state from given dict.

    :param state_d: a dictionary containing state data
    :return: a specialized instance of State
    """
    name = state_d['name']  # type: str
    stype = state_d.get('type', None)

    on_entry = state_d.get('on entry', None)  # type: Optional[str]
    on_entry = on_entry.strip() if on_entry else None
    on_exit = state_d.get('on exit', None)  # type: Optional[str]
    on_exit = on_exit.strip() if on_exit else None

    state = None  # type: Any

    if stype == 'final':
        state = FinalState(name, on_entry=on_entry, on_exit=on_exit)
    elif stype == 'shallow history':
        state = ShallowHistoryState(name, on_entry=on_entry, on_exit=on_exit,
                                    memory=state_d.get('memory', None))
    elif stype == 'deep history':
        state = DeepHistoryState(name, on_entry=on_entry, on_exit=on_exit,
                                 memory=state_d.get('memory', None))
    elif stype is None:
        substates = state_d.get('states', None)
        parallel_substates = state_d.get('parallel states', None)

        if substates and parallel_substates:
            raise StatechartError(
                '{} cannot declare both a "states" and a "parallel states" property'.format(name))
        elif substates:
            state = CompoundState(name, initial=state_d.get('initial', None),
                                  on_entry=on_entry, on_exit=on_exit)
        elif parallel_substates:
            state = OrthogonalState(name, on_entry=on_entry, on_exit=on_exit)
        else:
            state = BasicState(name, on_entry=on_entry, on_exit=on_exit)
    else:
        raise StatechartError('Unknown type {} for state {}'.format(stype, name))

    # Preconditions, postconditions and invariants
    for condition in state_d.get('contract', []):
        if condition.get('before', None):
            state.preconditions.append(condition['before'].strip())
        elif condition.get('after', None):
            state.postconditions.append(condition['after'].strip())
        elif condition.get('always', None):
            state.invariants.append(condition['always'].strip())
    return state


def export_to_dict(statechart: Statechart) -> Mapping[str, Any]:
    """
    Export given StateChart instance to a dict.

    :param statechart: a StateChart instance
    :return: a dict that can be used in *_import_from_dict*
    """
    d = dict()
    d['name'] = statechart.name
    if statechart.description:
        d['description'] = statechart.description
    if statechart.preamble:
        d['preamble'] = statechart.preamble

    d['root state'] = _export_state_to_dict(statechart, cast(str, statechart.root))

    return {'statechart': d}


def _export_state_to_dict(statechart: Statechart, state_name: str) -> Mapping[str, Any]:
    data = dict()

    state = statechart.state_for(state_name)

    data['name'] = state.name
    if isinstance(state, ShallowHistoryState):
        data['type'] = 'shallow history'
        if state.memory:
            data['memory'] = state.memory
    elif isinstance(state, DeepHistoryState):
        data['type'] = 'deep history'
        if state.memory:
            data['memory'] = state.memory
    elif isinstance(state, FinalState):
        data['type'] = 'final'

    if isinstance(state, ActionStateMixin):
        if state.on_entry:
            data['on entry'] = state.on_entry
        if state.on_exit:
            data['on exit'] = state.on_exit

    if isinstance(state, CompoundState):
        if state.initial:
            data['initial'] = state.initial

    preconditions = getattr(state, 'preconditions', [])
    postconditions = getattr(state, 'postconditions', [])
    invariants = getattr(state, 'invariants', [])
    if preconditions or postconditions or invariants:
        conditions = []
        for condition in preconditions:
            conditions.append({'before': condition})
        for condition in postconditions:
            conditions.append({'after': condition})
        for condition in invariants:
            conditions.append({'always': condition})
        data['contract'] = conditions

    if isinstance(state, TransitionStateMixin):
        # event, guard, target, action
        transitions = statechart.transitions_from(cast(StateMixin, state).name)
        if len(transitions) > 0:
            data['transitions'] = []

            for transition in transitions:
                transition_data = dict()
                if transition.event:
                    transition_data['event'] = transition.event
                if transition.guard:
                    transition_data['guard'] = transition.guard
                if transition.target:
                    transition_data['target'] = transition.target
                if transition.action:
                    transition_data['action'] = transition.action
                if transition.priority != Transition.DEFAULT_PRIORITY:
                    if transition.priority == Transition.LOW_PRIORITY:
                        priority = 'low'
                    elif transition.priority == Transition.HIGH_PRIORITY:
                        priority = 'high'
                    else:
                        priority = transition.priority
                    transition_data['priority'] = priority

                preconditions = getattr(transition, 'preconditions', [])
                postconditions = getattr(transition, 'postconditions', [])
                invariants = getattr(transition, 'invariants', [])
                if preconditions or postconditions or invariants:
                    conditions = []
                    for condition in preconditions:
                        conditions.append({'before': condition})
                    for condition in postconditions:
                        conditions.append({'after': condition})
                    for condition in invariants:
                        conditions.append({'always': condition})
                    transition_data['contract'] = conditions

                data['transitions'].append(transition_data)

    if isinstance(state, CompositeStateMixin):
        children = statechart.children_for(cast(StateMixin, state).name)
        children_data = [_export_state_to_dict(statechart, child) for child in children]

        if isinstance(state, CompoundState):
            data['states'] = children_data
        elif isinstance(state, OrthogonalState):
            data['parallel states'] = children_data

    return data



from collections import defaultdict


def sorted_groupby(iterable, key=None, reverse=False):
    """
    Return pairs (label, group) grouped and sorted by label = key(item).
    """
    groups = defaultdict(list)
    for value in iterable:
        groups[key(value)].append(value)

    def sort_key(e):
        return e[0]

    return sorted(groups.items(), key=sort_key, reverse=reverse)



from typing import Union, Optional, List, Any, Mapping
from .interpreter import Interpreter
from .model import MacroStep, Transition


__all__ = [
    'state_is_entered', 'state_is_exited',
    'event_is_fired', 'event_is_consumed',
    'transition_is_processed',
    'expression_holds',
]

MacroSteps = Union[MacroStep, List[MacroStep]]


def state_is_entered(steps: MacroSteps, name: str) -> bool:
    """
    Holds if state was entered during given steps.

    :param steps: a macrostep or list of macrosteps
    :param name: name of a state
    :return: given state was entered
    """
    steps = steps if isinstance(steps, list) else [steps]
    for step in steps:
        if name in step.entered_states:
            return True
    return False


def state_is_exited(steps: MacroSteps, name: str) -> bool:
    """
    Holds if state was exited during given steps.

    :param steps: a macrostep or list of macrosteps
    :param name: name of a state
    :return: given state was exited
    """
    steps = steps if isinstance(steps, list) else [steps]
    for step in steps:
        if name in step.exited_states:
            return True
    return False


def event_is_fired(
        steps: MacroSteps, name: Optional[str],
        parameters: Mapping[str, Any] = None) -> bool:
    """
    Holds if an event was fired during given steps.

    If name is None, this function looks for any event.
    If parameters are provided, their values are compared with the respective
    attribute of the event. Not *all* parameters have to be provided, as only
    the ones that are provided are actually compared.

    :param steps: a macrostep or list of macrosteps
    :param name: name of an event
    :param parameters: additional parameters
    :return: event was fired
    """
    steps = steps if isinstance(steps, list) else [steps]
    parameters = dict() if parameters is None else parameters

    for step in steps:
        for event in step.sent_events:
            if name is None or event.name == name:
                matching_parameters = True
                for key, value in parameters.items():
                    if getattr(event, key, None) != value:
                        matching_parameters = False
                        break
                if matching_parameters:
                    return True
    return False


def event_is_consumed(
        steps: MacroSteps, name: Optional[str],
        parameters: Mapping[str, Any] = None) -> bool:
    """
    Holds if an event was consumed during given steps.

    If name is None, this function looks for any event.
    If parameters are provided, their values are compared with the respective
    attribute of the event. Not *all* parameters have to be provided, as only
    the ones that are provided are actually compared.

    :param steps: a macrostep or list of macrosteps
    :param name: name of an event
    :param parameters: additional parameters
    :return: event was consumed
    """
    steps = steps if isinstance(steps, list) else [steps]
    parameters = dict() if parameters is None else parameters

    for step in steps:
        if step.event is None:
            continue

        if name is None or step.event.name == name:
            matching_parameters = True
            for key, value in parameters.items():
                if getattr(step.event, key, None) != value:
                    matching_parameters = False
                    break
            if matching_parameters:
                return True
    return False


def transition_is_processed(steps: MacroSteps, transition: Optional[Transition] = None) -> bool:
    """
    Holds if a transition was processed during given steps.

    If no transition is provided, this function looks for any transition.

    :param steps: a macrostep or list of macrosteps
    :param transition: a transition
    :return: transition was processed
    """
    steps = steps if isinstance(steps, list) else [steps]

    if transition is None:
        for step in steps:
            if len(step.transitions) > 0:
                return True
        return False
    else:
        for step in steps:
            if transition in step.transitions:
                return True
        return False


def expression_holds(interpreter: Interpreter, expression: str) -> bool:
    """
    Holds if given expression holds.

    :param interpreter: current interpreter
    :param expression: expression to evaluate
    :return: expression holds
    """
    return interpreter._evaluator._evaluate_code(expression)



__package__ = 'sismic'
__version__ = '1.6.8'
__licence__ = 'LGPL3'
__author__ = 'Alexandre Decan'
__url__ = 'https://github.com/AlexandreDecan/sismic/'
__description__ = 'Sismic Interactive Statechart Model Interpreter and Checker'



import pytest
import os

from subprocess import check_call

from sismic.interpreter import Event
from sismic import testing


def test_writer(writer):
    writer.queue(
        Event('keyPress', key='bonjour '),
        Event('toggle'),
        Event('keyPress', key='a '),
        Event('toggle'),
        Event('toggle_bold'),
        Event('keyPress', key='tous !'),
        Event('leave')
    )

    writer.execute()
    assert writer.final
    assert writer.context['output'] == ['bonjour ', '[b]', '[i]', 'a ', '[/b]', '[/i]', '[b]', 'tous !', '[/b]']


class TestElevator:
    def test_init(self, elevator):
        assert elevator.configuration == []

        # Stabilisation
        elevator.execute_once()

        assert set(elevator.configuration) == {'active', 'floorListener', 'floorSelecting',
                                               'movingElevator', 'doorsOpen'}

    def test_floor_selection(self, elevator):
        # Stabilisation
        elevator.execute_once()

        elevator.queue('floorSelected', floor=4).execute_once()
        assert elevator.context['destination'] == 4

        elevator.execute_once()
        assert set(elevator.configuration) == {'active', 'active', 'doorsClosed', 'floorListener', 'floorSelecting', 'movingElevator'}

    def test_floor_selected_and_reached(self, elevator):
        # Stabilisation
        elevator.execute_once()

        elevator.queue('floorSelected', floor=4).execute()
        assert elevator.context['current'] == 4

        elevator.clock.time += 10
        elevator.execute()

        assert 'doorsOpen' in elevator.configuration
        assert elevator.context['current'] == 0


class TestRemoteElevator:
    def test_button(self, elevator, remote_elevator):
        assert elevator.context['current'] == 0

        steps = remote_elevator.queue('button_2_pushed').execute()
        assert testing.event_is_fired(steps, 'floorSelected', {'floor': 2})

        elevator.execute()
        assert elevator.context['current'] == 2

    def test_button_0_on_groundfloor(self, elevator, remote_elevator):
        assert elevator.context['current'] == 0

        remote_elevator.queue('button_0_pushed').execute()
        elevator.execute()

        assert elevator.context['current'] == 0


class TestMicrowave:
    def test_lamp_on(self, microwave):
        microwave.execute_once()
        microwave.queue('door_opened')

        steps = microwave.execute_once()
        assert testing.event_is_fired(steps, 'lamp_switch_on')

    def test_heating_on(self, microwave):
        microwave.execute_once()
        microwave.queue('door_opened', 'item_placed', 'door_closed', 'timer_inc').execute()

        microwave.queue('cooking_start')
        steps = microwave.execute()
        assert testing.event_is_fired(steps, 'heating_on')
        assert testing.event_is_fired(steps, 'lamp_switch_on')
        assert testing.event_is_fired(steps, 'turntable_start')



class Stopwatch:
    def __init__(self):
        self.elapsed_time = 0
        self.split_time = 0
        self.is_split = False
        self.running = False

    def start(self):
        # Start internal timer
        self.running = True

    def stop(self):
        # Stop internal timer
        self.running = False

    def reset(self):
        # Reset internal timer
        self.elapsed_time = 0

    def split(self):
        # Split time
        if not self.is_split:
            self.is_split = True
            self.split_time = self.elapsed_time

    def unsplit(self):
        # Unsplit time
        if self.is_split:
            self.is_split = False

    def display(self):
        # Return the value to display
        if self.is_split:
            return int(self.split_time)
        else:
            return int(self.elapsed_time)

    def update(self, delta):
        # Update internal timer of ``delta`` seconds
        if self.running:
            self.elapsed_time += delta



import time
import tkinter as tk

from sismic.interpreter import Interpreter
from sismic.io import import_from_yaml
from stopwatch import Stopwatch


# The two following lines are NOT needed in a typical environment.
# These lines make sismic available in our testing environment
import sys
sys.path.append('../../..')




# Create a tiny GUI
class StopwatchApplication(tk.Frame):
    def __init__(self, master=None):
        super().__init__(master)

        # Initialize widgets
        self.create_widgets()

        # Create a Stopwatch interpreter
        with open('stopwatch_external.yaml') as f:
            statechart = import_from_yaml(f)

        # Create a stopwatch object and pass it to the interpreter
        self.stopwatch = Stopwatch()
        self.interpreter = Interpreter(statechart, initial_context={'stopwatch': self.stopwatch})
        self.interpreter.clock.start()
        
        # Run the interpreter
        self.run()

        # Update the stopwatch every 100ms
        self.after(100, self.update_stopwatch)

    def update_stopwatch(self):
        self.stopwatch.update(delta=0.1)
        self.after(100, self.update_stopwatch)

        # Update timer label
        self.w_timer['text'] = self.stopwatch.display()

    def run(self):
        # Queue a call every 100ms on tk's mainloop
        self.interpreter.execute()
        self.after(100, self.run)
        self.w_states['text'] = 'active states: ' + ', '.join(self.interpreter.configuration)

    def create_widgets(self):
        self.pack()

        # Add buttons
        self.w_btn_start = tk.Button(self, text='start', command=self._start)
        self.w_btn_stop = tk.Button(self, text='stop', command=self._stop)
        self.w_btn_split = tk.Button(self, text='split', command=self._split)
        self.w_btn_unsplit = tk.Button(self, text='unsplit', command=self._unsplit)
        self.w_btn_reset = tk.Button(self, text='reset', command=self._reset)
        self.w_btn_quit = tk.Button(self, text='quit', command=self._quit)

        # Initial button states
        self.w_btn_stop['state'] = tk.DISABLED
        self.w_btn_unsplit['state'] = tk.DISABLED

        # Pack
        self.w_btn_start.pack(side=tk.LEFT,)
        self.w_btn_stop.pack(side=tk.LEFT,)
        self.w_btn_split.pack(side=tk.LEFT,)
        self.w_btn_unsplit.pack(side=tk.LEFT,)
        self.w_btn_reset.pack(side=tk.LEFT,)
        self.w_btn_quit.pack(side=tk.LEFT,)

        # Active states label
        self.w_states = tk.Label(root)
        self.w_states.pack(side=tk.BOTTOM, fill=tk.X)

        # Timer label
        self.w_timer = tk.Label(root, font=("Helvetica", 16), pady=5)
        self.w_timer.pack(side=tk.BOTTOM, fill=tk.X)

    def _start(self):
        self.interpreter.queue('start')
        self.w_btn_start['state'] = tk.DISABLED
        self.w_btn_stop['state'] = tk.NORMAL

    def _stop(self):
        self.interpreter.queue('stop')
        self.w_btn_start['state'] = tk.NORMAL
        self.w_btn_stop['state'] = tk.DISABLED

    def _reset(self):
        self.interpreter.queue('reset')

    def _split(self):
        self.interpreter.queue('split')
        self.w_btn_split['state'] = tk.DISABLED
        self.w_btn_unsplit['state'] = tk.NORMAL

    def _unsplit(self):
        self.interpreter.queue('split')
        self.w_btn_split['state'] = tk.NORMAL
        self.w_btn_unsplit['state'] = tk.DISABLED

    def _quit(self):
        self.master.destroy()


if __name__ == '__main__':
    # Create GUI
    root = tk.Tk()
    root.wm_title('StopWatch (external)')
    app = StopwatchApplication(master=root)

    app.mainloop()



import abc

from time import time


__all__ = ['Clock', 'SimulatedClock', 'UtcClock', 'SynchronizedClock']


class Clock(metaclass=abc.ABCMeta):
    """
    Abstract implementation of a clock, as used by an interpreter.

    The purpose of a clock instance is to provide a way for the interpreter
    to get the current time during the execution of a statechart.
    """
    @abc.abstractproperty
    def time(self) -> float:
        """
        Current time
        """
        raise NotImplementedError()

    def __repr__(self):
        return '{}[{}]'.format(self.__class__.__name__, self.time)


class SimulatedClock(Clock):
    """
    A simulated clock, starting from 0, that can be manually or automatically
    incremented.

    Manual incrementation can be done by setting a new value to the time attribute.
    Automatic incrementation occurs when start() is called, until stop() is called.
    In that case, clock speed can be adjusted with the speed attribute.
    A value strictly greater than 1 increases clock speed while a value strictly
    lower than 1 slows down the clock.
    """

    def __init__(self) -> None:
        self._base = time()
        self._time = 0
        self._play = False
        self._speed = 1

    @property
    def _elapsed(self):
        return (time() - self._base) * self._speed if self._play else 0

    def start(self) -> None:
        """
        Clock will be automatically updated both based on real time and
        its speed attribute.
        """
        if not self._play:
            self._base = time()
            self._play = True

    def stop(self) -> None:
        """
        Clock won't be automatically updated.
        """
        if self._play:
            self._time += self._elapsed
            self._play = False

    @property
    def speed(self) -> float:
        """
        Speed of the current clock. Only affects time if start() is called.
        """
        return self._speed

    @speed.setter
    def speed(self, speed):
        self._time += self._elapsed
        self._base = time()
        self._speed = speed

    @property
    def time(self) -> float:
        """
        Time value of this clock.
        """
        return self._time + self._elapsed

    @time.setter
    def time(self, new_time):
        """
        Set the time of this clock.

        :param new_time: new time
        """
        current_time = self.time
        if new_time < current_time:
            raise ValueError('Time must be monotonic, cannot change time from {} to {}'.format(
                current_time, new_time))

        self._time = new_time
        self._base = time()

    def __str__(self):
        return '{:.2f}'.format(float(self.time))

    def __repr__(self):
        return '{}[{:.2f},x{},{}]'.format(
            self.__class__.__name__,
            self.time,
            self._speed,
            '>' if self._play else '=',
        )


class UtcClock(Clock):
    """
    A clock that simulates a wall clock in UTC.

    The returned time value is based on Python time.time() function.
    """

    @property
    def time(self) -> float:
        return time()


class SynchronizedClock(Clock):
    """
    A clock that is synchronized with a given interpreter.

    The synchronization is based on the interpreter's internal time value, not
    on its clock. As a consequence, the time value of a SynchronizedClock only
    changes when the underlying interpreter is executed.

    :param interpreter: an interpreter instance
    """

    def __init__(self, interpreter) -> None:
        self._interpreter = interpreter

    @property
    def time(self) -> float:
        return self._interpreter.time



import bisect
import warnings

from itertools import combinations
from typing import (Any, Callable, Dict, Iterable, List, Mapping, Optional,
                    Set, Tuple, Union, cast)

from .listener import InternalEventListener, PropertyStatechartListener
from ..utilities import sorted_groupby
from ..clock import Clock, SimulatedClock, SynchronizedClock
from ..code import Evaluator, PythonEvaluator
from ..exceptions import (ConflictingTransitionsError, InvariantError,
                          NonDeterminismError, PostconditionError,
                          PreconditionError)
from ..model import (CompoundState, DeepHistoryState, Event,
                     FinalState, InternalEvent, MacroStep, MetaEvent,
                     MicroStep, OrthogonalState, ShallowHistoryState,
                     Statechart, StateMixin, Transition)

__all__ = ['Interpreter']


class _KeyifyList():
    def __init__(self, inner, key):
        self.inner = inner
        self.key = key

    def __len__(self):
        return len(self.inner)

    def __getitem__(self, k):
        return self.key(self.inner[k])


class Interpreter:
    """
    A discrete interpreter that executes a statechart according to a semantic close to SCXML
    (eventless transitions first, inner-first/source state semantics).

    :param statechart: statechart to interpret
    :param evaluator_klass: An optional callable (e.g. a class) that takes an interpreter and an
        optional initial context as input and returns an *Evaluator* instance that will be used to
        initialize the interpreter. By default, the *PythonEvaluator* class will be used.
    :param initial_context: an optional initial context that will be provided to the evaluator.
        By default, an empty context is provided
    :param clock: A BaseClock instance that will be used to set this interpreter internal time.
        By default, a SimulatedClock is used.
    :param ignore_contract: set to True to ignore contract checking during the execution.
    """

    def __init__(self, statechart: Statechart, *,
                 evaluator_klass: Callable[..., Evaluator] = PythonEvaluator,
                 initial_context: Mapping[str, Any] = None,
                 clock: Clock = None,
                 ignore_contract: bool = False) -> None:
        # Internal variables
        self._ignore_contract = ignore_contract
        self._statechart = statechart

        self._initialized = False

        # Internal clock
        self.clock = SimulatedClock() if clock is None else clock
        self._time = self.clock.time

        # History states memory
        self._memory = {}  # type: Dict[str, Optional[List[str]]]

        # Set of active states
        self._configuration = set()  # type: Set[str]

        # Entry and idle times
        self._entry_time = dict()  # type: Dict[str, float]
        self._idle_time = dict()  # type: Dict[str, float]

        # Events sent during current macro step
        self._sent_events = []  # type: List[Event]

        # Event queues
        self._internal_queue = []  # type: List[Tuple[float, InternalEvent]]
        self._external_queue = []  # type: List[Tuple[float, Event]]

        # Bound listeners
        self._listeners = []  # type: List[Callable[[MetaEvent], Any]]

        # Evaluator
        self._evaluator = evaluator_klass(self, initial_context=initial_context)
        self._evaluator.execute_statechart(statechart)

    @property
    def time(self) -> float:
        """
        Time of the latest execution.
        """
        return self._time

    @time.setter
    def time(self, value: float):
        warnings.warn(
            'Interpreter.time is deprecated since 1.3.0, use Interpreter.clock.time instead',
            DeprecationWarning)
        self.clock.time = value  # type: ignore

    @property
    def configuration(self) -> List[str]:
        """
        List of active states names, ordered by depth. Ties are broken according to the
        lexicographic order on the state name.
        """
        return sorted(self._configuration, key=lambda s: (self._statechart.depth_for(s), s))

    @property
    def context(self) -> Mapping[str, Any]:
        """
        The context of execution.
        """
        return self._evaluator.context

    @property
    def final(self) -> bool:
        """
        Boolean indicating whether this interpreter is in a final configuration.
        """
        return self._initialized and len(self._configuration) == 0

    @property
    def statechart(self) -> Statechart:
        """
        Embedded statechart
        """
        return self._statechart

    def attach(self, listener: Callable[[MetaEvent], Any]) -> None:
        """
        Attach given listener to the current interpreter.

        The listener is called each time a meta-event is emitted by current interpreter.
        Emitted meta-events are:

        - *step started*: when a (possibly empty) macro step starts. The current time of the step
          is available through the ``time`` attribute.
        - *step ended*: when a (possibly empty) macro step ends.
        - *event consumed*: when an event is consumed. The consumed event is exposed through the
          ``event`` attribute.
        - *event sent*: when an event is sent. The sent event is exposed through the ``event``
          attribute.
        - *state exited*: when a state is exited. The exited state is exposed through the ``state``
          attribute.
        - *state entered*: when a state is entered. The entered state is exposed through the
          ``state`` attribute.
        - *transition processed*: when a transition is processed. The source state, target state
          and the event ard exposed respectively through the ``source``, ``target`` and ``event``
          attributes.
        - Every meta-event that is sent from within the statechart.

        This is a low-level interface for ``self.bind`` and ``self.bind_property_statechart``.

        Consult ``sismic.interpreter.listener`` for common listeners/wrappers.

        :param listener: A callable that accepts meta-event instances.
        """
        self._listeners.append(listener)

    def detach(self, listener: Callable[[MetaEvent], Any]) -> None:
        """
        Remove given listener from the ones that are currently attached to this interpreter.

        :param listener: A previously attached listener.
        """
        self._listeners.remove(listener)

    def bind(self, interpreter_or_callable: Union['Interpreter', Callable[[
             Event], Any]]) -> Callable[[MetaEvent], Any]:
        """
        Bind an interpreter (or a callable) to the current interpreter.

        Internal events sent by this interpreter will be propagated as external events.
        If *interpreter_or_callable* is an *Interpreter* instance,  its *queue* method is called.
        This is, if *i1* and *i2* are interpreters, *i1.bind(i2)* is equivalent to
        *i1.bind(i2.queue)*.

        This method is a higher-level interface for ``self.attach``.
        If ``x = interpreter.bind(...)``, use ``interpreter.detach(x)`` to unbind a
        previously bound interpreter.

        :param interpreter_or_callable: interpreter or callable to bind.
        :return: the resulting attached listener.
        """
        if isinstance(interpreter_or_callable, Interpreter):
            listener = InternalEventListener(interpreter_or_callable.queue)
        else:
            listener = InternalEventListener(interpreter_or_callable)

        self.attach(listener)

        return listener

    def bind_property_statechart(
            self, statechart: Statechart, *, interpreter_klass: Callable = None) -> Callable[
            [MetaEvent],
            Any]:
        """
        Bind a property statechart to the current interpreter.

        A property statechart receives meta-events from the current interpreter depending on
        what happens. See ``attach`` method for a full list of meta-events.

        The internal clock of all property statecharts is synced with the one of the current
        interpreter. As soon as a property statechart reaches a final state, a
        ``PropertyStatechartError`` will be raised, meaning that the property expressed by the
        corresponding property statechart is not satisfied. Property statecharts are automatically
        executed when they are bound to an interpreter.

        Since Sismic 1.4.0: passing an interpreter as first argument is deprecated.

        This method is a higher-level interface for ``self.attach``.
        If ``x = interpreter.bind_property_statechart(...)``, use ``interpreter.detach(x)`` to
        unbind a previously bound property statechart.

        :param statechart: A statechart instance.
        :param interpreter_klass: An optional callable that accepts a statechart as first parameter
            and a named parameter clock. Default to Interpreter.
        :return: the resulting attached listener.
        """
        if isinstance(statechart, Interpreter):
            warnings.warn(
                'Passing an interpreter to bind_property_statechart is deprecated since 1.4.0. '
                'Use interpreter_klass instead.',
                DeprecationWarning)
            interpreter = statechart
            interpreter.clock = SynchronizedClock(self)
        else:
            interpreter_klass = Interpreter if interpreter_klass is None else interpreter_klass
            interpreter = interpreter_klass(statechart, clock=SynchronizedClock(self))

        listener = PropertyStatechartListener(interpreter)
        self.attach(listener)

        return listener

    def queue(self, event_or_name: Union[str, Event],
              *event_or_names: Union[str, Event],
              **parameters) -> 'Interpreter':
        """
        Create and queue given events to the external event queue.

        If an event has a `delay` parameter, it will be processed by the first call to
        `execute_once` as soon as `self.clock.time` exceeds current `self.time + event.delay`.

        If named parameters are provided, they will be added to all events
        that are provided by name.

        :param event_or_name: name of the event or Event instance
        :param event_or_names: additional events
        :param parameters: event parameters.
        :return: *self* so it can be chained.
        """
        for event in [event_or_name] + list(event_or_names):
            event = Event(event, **parameters) if isinstance(event, str) else event
            self._queue_event(event)
        return self

    def execute(self, max_steps: int = -1) -> List[MacroStep]:
        """
        Repeatedly calls *execute_once* and return a list containing
        the returned values of *execute_once*.

        Notice that this does NOT return an iterator but computes the whole list first
        before returning it.

        :param max_steps: An upper bound on the number steps that are computed and returned.
            Default is -1, no limit. Set to a positive integer to avoid infinite loops
            in the statechart execution.
        :return: A list of *MacroStep* instances
        """
        returned_steps = []
        i = 0
        macro_step = self.execute_once()
        while macro_step:
            returned_steps.append(macro_step)
            i += 1
            if 0 < max_steps == i:
                break
            macro_step = self.execute_once()
        return returned_steps

    def execute_once(self) -> Optional[MacroStep]:
        """
        Select transitions that can be fired based on available queued events, process them and
        stabilize the interpreter. When multiple transitions are selected, they are atomically
        processed: states are exited, transition is processed, states are entered, statechart is
        stabilized and only after that, the next transition is processed.

        :return: a macro step or *None* if nothing happened
        """
        # Store time to have a consistent time value during this step
        self._time = self.clock.time

        # Reset the list of events that were sent
        self._sent_events.clear()

        # Notify listeners
        self._raise_event(MetaEvent('step started', time=self.time))

        # Compute steps
        computed_steps = self._compute_steps()

        if len(computed_steps) > 0:

            # Consume event if it triggered a transition
            if computed_steps[0].event is not None:
                event = self._select_event(consume=True)
                self._raise_event(MetaEvent('event consumed', event=event))
            else:
                event = None

            # Execute the steps
            if hasattr(self._evaluator, 'on_step_starts'):
                warnings.warn('Evaluator.on_step_starts is deprecated since 1.4.0.',
                              DeprecationWarning)
                self._evaluator.on_step_starts(event)

            executed_steps = []
            for step in computed_steps:
                executed_steps.append(self._apply_step(step))
                executed_steps.extend(self._stabilize())

            # type: Optional[MacroStep]
            macro_step = MacroStep(time=self.time, steps=executed_steps)
        else:  # No step
            macro_step = None

        # Check state invariants
        configuration = self.configuration  # Use self.configuration to benefit from the sorting
        for name in configuration:
            state = self._statechart.state_for(name)
            self._evaluate_contract_conditions(state, 'invariants', macro_step)

        self._raise_event(MetaEvent('step ended'))

        return macro_step

    def _queue_event(self, event: Event):
        """
        Convenient helper to queue events wrt. to internal/external and their (optional) delay.

        :param event: Event to queue.
        """
        if isinstance(event, InternalEvent):
            queue = cast(List[Tuple[float, Event]], self._internal_queue)
        else:
            queue = self._external_queue

        time = self.time + getattr(event, 'delay', 0)
        position = bisect.bisect_right(  # type: ignore
            _KeyifyList(queue, lambda t: (t[0], not isinstance(t[1], InternalEvent))),
            (time, not isinstance(event, InternalEvent))
        )
        queue.insert(position, (time, event))

    def _raise_event(self, event: Union[InternalEvent, MetaEvent]) -> None:
        """
        Raise an event from the statechart.

        Only InternalEvent and MetaEvent (and their subclasses) are accepted.

        :param event: event to be sent by the statechart.
        """
        if isinstance(event, InternalEvent):
            self._queue_event(event)
            self._raise_event(MetaEvent('event sent', event=event))
            if hasattr(event, 'delay'):
                # Deprecated since 1.4.0
                self._raise_event(MetaEvent('delayed event sent', event=event))
        elif isinstance(event, MetaEvent):
            for listener in self._listeners:
                listener(event)
        else:
            raise ValueError(
                'Only InternalEvent and MetaEvent can be sent by a statechart, not {}'.format(
                    type(event)))

    def _select_event(self, *, consume: bool = False) -> Optional[Event]:
        """
        Return the next event to process.
        Internal events have priority over external ones.

        :param consume: Indicates whether event should be consumed, default to False.
        :return: An instance of Event or None if no event is available
        """
        for queue in cast(
                Tuple[List[Tuple[float, Event]]],
                (self._internal_queue, self._external_queue)):
            if len(queue) > 0:
                time, event = queue[0]
                if time <= self.time:
                    if consume:
                        queue.pop(0)
                    return event
        return None

    def _select_transitions(self, event: Optional[Event], states: Iterable[str], *,
                            eventless_first=True, inner_first=True) -> List[Transition]:
        """
        Select and return the transitions that are triggered, based on given event
        (or None if no event can be consumed) and given list of states.

        By default, this function prioritizes eventless transitions and follows
        inner-first/source state semantics.

        :param event: event to consider, possibly None.
        :param states: state names to consider.
        :param eventless_first: True to prioritize eventless transitions.
        :param inner_first: True to follow inner-first/source state semantics.
        :return: list of triggered transitions.
        """
        selected_transitions = []  # type: List[Transition]
        considered_transitions = []  # type: List[Transition]
        _state_depth_cache = dict()  # type: Dict[str, int]

        # Select triggerable (based on event) transitions for considered states
        for transition in self._statechart.transitions:
            if transition.source in states:
                if transition.event is None or transition.event == getattr(event, 'name', None):
                    # Compute order based on depth
                    if transition.source not in _state_depth_cache:
                        _state_depth_cache[transition.source] = self._statechart.depth_for(
                            transition.source)

                    considered_transitions.append(transition)

        # Which states should be selected to satisfy depth ordering?
        if inner_first:
            ignored_state_selector = self._statechart.ancestors_for
        else:
            ignored_state_selector = self._statechart.descendants_for
        ignored_states = set()  # type: Set[str]

        # Group and sort transitions based on the event
        def eventless_first_order(t):
            return t.event is not None

        for has_event, transitions in sorted_groupby(
                considered_transitions, key=eventless_first_order, reverse=not eventless_first):
            # If there are selected transitions (from previous group), ignore new ones
            if len(selected_transitions) > 0:
                break

            # Event shouldn't be exposed to guards if we're processing eventless transition
            exposed_event = event if has_event else None

            # Group and sort transitions based on the source state depth
            def depth_order(t):
                return _state_depth_cache[t.source]

            for _, transitions in sorted_groupby(transitions, key=depth_order, reverse=inner_first):
                # Group and sort transitions based on the source state
                def state_order(t):
                    return t.source  # we just want states to be grouped here

                for source, transitions in sorted_groupby(transitions, key=state_order):
                    # Do not considered ignored states
                    if source in ignored_states:
                        continue

                    has_found_transitions = False

                    # Group and sort transitions based on their priority
                    def priority_order(t):
                        return t.priority

                    for _, transitions in sorted_groupby(
                            transitions, key=priority_order, reverse=True):
                        for transition in transitions:
                            if transition.guard is None or self._evaluator.evaluate_guard(
                                    transition, exposed_event):
                                # Add transition to the list of selected ones
                                selected_transitions.append(transition)
                                has_found_transitions = True

                        # Ignore ancestors/descendants w.r.t. inner-first/source state
                        if has_found_transitions:
                            for state in ignored_state_selector(source):
                                ignored_states.add(state)
                            # Also ignore current state, as we found transitions in a higher
                            # priority class
                            ignored_states.add(source)
                            break

        return selected_transitions

    def _sort_transitions(self, transitions: List[Transition]) -> List[Transition]:
        """
        Given a list of triggered transitions, return a list of transitions in an order that
        represents the order in which they have to be processed.

        :param transitions: a list of *Transition* instances
        :return: an ordered list of *Transition* instances
        :raise ExecutionError: In case of non-determinism (*NonDeterminismError*) or conflicting
            transitions (*ConflictingTransitionsError*).
        """
        if len(transitions) > 1:
            # If more than one transition, we check (1) they are from separate regions and (2) they
            # do not conflict. Two transitions conflict if one of them leaves the parallel state
            for t1, t2 in combinations(transitions, 2):
                # Check (1)
                lca = cast(str, self._statechart.least_common_ancestor(t1.source, t2.source))
                lca_state = self._statechart.state_for(lca)

                # Their LCA must be an orthogonal state!
                if not isinstance(lca_state, OrthogonalState):
                    raise NonDeterminismError(
                        'Non-determinist choice between transitions {t1} and {t2}'
                        '\nConfiguration is {c}\nEvent is {e}\nTransitions are:{t}\n'
                        .format(c=self.configuration, e=t1.event, t=transitions, t1=t1, t2=t2)
                    )

                # Check (2)
                # This check must be done wrt. to LCA, as the combination of from_states could
                # come from nested parallel regions!
                for transition in [t1, t2]:
                    last_before_lca = transition.source
                    for state in self._statechart.ancestors_for(transition.source):
                        if state == lca:
                            break
                        last_before_lca = state
                    # Target must be a descendant (or self) of this state
                    if (transition.target and (transition.target not in [
                            last_before_lca] + self._statechart.descendants_for(last_before_lca))):
                        raise ConflictingTransitionsError(
                            'Conflicting transitions: {t1} and {t2}'
                            '\nConfiguration is {c}\nEvent is {e}\nTransitions are:{t}\n'
                            .format(c=self.configuration, e=t1.event, t=transitions, t1=t1, t2=t2)
                        )

            # Define an arbitrary order based on the depth and the name of source states.
            transitions = sorted(
                transitions, key=lambda t: (-self._statechart.depth_for(t.source), t.source))

        return transitions

    def _compute_steps(self) -> List[MicroStep]:
        """
        Compute and returns the next steps based on current configuration
        and event queues.

        :return: a possibly empty list of steps
        """
        # Initialization
        if not self._initialized:
            self._initialized = True
            return [MicroStep(entered_states=[cast(str, self._statechart.root)])]

        # Select transitions
        event = self._select_event()
        transitions = self._select_transitions(event, states=self._configuration)

        # No transition can be triggered?
        if len(transitions) == 0:
            if event is None:
                # No event, no step!
                return []
            else:
                # Empty step, so that event is eventually consumed
                return [MicroStep(event=event)]

        # Compute transitions order
        transitions = self._sort_transitions(transitions)

        # Should the step consume an event?
        event = None if transitions[0].event is None else event

        return self._create_steps(event, transitions)

    def _create_steps(self, event: Optional[Event],
                      transitions: Iterable[Transition]) -> List[MicroStep]:
        """
        Return a (possibly empty) list of micro steps. Each micro step corresponds to the process
        of a transition matching given event.

        :param event: the event to consider, if any
        :param transitions: the transitions that should be processed
        :return: a list of micro steps.
        """
        returned_steps = []
        for transition in transitions:
            # Internal transition
            if transition.target is None:
                returned_steps.append(MicroStep(event=event, transition=transition))
                continue

            lca = self._statechart.least_common_ancestor(transition.source, transition.target)
            from_ancestors = self._statechart.ancestors_for(transition.source)
            to_ancestors = self._statechart.ancestors_for(transition.target)

            # Exited states
            exited_states = []

            # last_before_lca is the "highest" ancestor of from_state that is a child of LCA
            last_before_lca = transition.source
            for state in from_ancestors:
                if state == lca:
                    break
                last_before_lca = state

            # Take all the descendants of this state and list the ones that are active
            # Mind the reversed order!
            for descendant in self._statechart.descendants_for(last_before_lca)[::-1]:
                # Only leave states that are currently active
                if descendant in self._configuration:
                    exited_states.append(descendant)

            # Add last_before_lca as it is a child of LCA that must be exited
            if last_before_lca in self._configuration:
                exited_states.append(last_before_lca)

            # Entered states
            entered_states = [transition.target]
            for state in to_ancestors:
                if state == lca:
                    break
                entered_states.insert(0, state)

            returned_steps.append(
                MicroStep(
                    event=event, transition=transition, entered_states=entered_states,
                    exited_states=exited_states))

        return returned_steps

    def _create_stabilization_step(self, names: Iterable[str]) -> Optional[MicroStep]:
        """
        Return a stabilization step, ie. a step that lead to a more stable situation
        for the current statechart. Stabilization means:

         - Enter the initial state of a compound state with no active child
         - Enter the memory of a history state
         - Enter the children of an orthogonal state with no active child
         - Empty active configuration if root's child is a final state

        :param names: List of states to consider (usually, the active configuration)
        :return: A *MicroStep* instance or *None* if this statechart can not be more stabilized
        """
        # Check if we are in a set of "stable" states
        leaves_names = self._statechart.leaf_for(names)
        leaves = sorted([self._statechart.state_for(name) for name in leaves_names],
                        key=lambda s: (-self._statechart.depth_for(s.name), s.name))

        for leaf in leaves:
            if isinstance(
                    leaf, FinalState) and self._statechart.parent_for(
                    leaf.name) == self._statechart.root:
                return MicroStep(exited_states=[leaf.name, cast(str, self._statechart.root)])
            if isinstance(leaf, (ShallowHistoryState, DeepHistoryState)):
                states_to_enter = cast(List[str], self._memory.get(leaf.name, [leaf.memory]))
                states_to_enter.sort(key=lambda x: (self._statechart.depth_for(x), x))
                return MicroStep(entered_states=states_to_enter, exited_states=[leaf.name])
            elif isinstance(leaf, OrthogonalState) and self._statechart.children_for(leaf.name):
                return MicroStep(entered_states=sorted(self._statechart.children_for(leaf.name)))
            elif isinstance(leaf, CompoundState) and leaf.initial:
                return MicroStep(entered_states=[leaf.initial])

        return None

    def _apply_step(self, step: MicroStep) -> MicroStep:
        """
        Apply given *MicroStep* on this statechart

        :param step: *MicroStep* instance
        :return: a new MicroStep, completed with sent events
        """
        entered_states = list(map(self._statechart.state_for, step.entered_states))
        exited_states = list(map(self._statechart.state_for, step.exited_states))

        active_configuration = set(self._configuration)  # Copy

        sent_events = []  # type: List[Event]

        # Exit states
        for state in exited_states:
            # Execute exit action
            sent_events.extend(self._evaluator.execute_on_exit(state))

            # Deal with history
            if isinstance(state, CompoundState):
                # Look for an HistoryStateMixin among its children
                for child_name in self._statechart.children_for(state.name):
                    child = self._statechart.state_for(child_name)
                    if isinstance(child, DeepHistoryState):
                        # This MUST contain at least one element!
                        active = active_configuration.intersection(
                            self._statechart.descendants_for(state.name))
                        assert len(active) >= 1
                        self._memory[child.name] = list(active)
                    elif isinstance(child, ShallowHistoryState):
                        # This MUST contain exactly one element!
                        active = active_configuration.intersection(
                            self.statechart.children_for(state.name))
                        assert len(active) == 1
                        self._memory[child.name] = list(active)

            # Remove state from active configuration
            self._configuration.remove(state.name)

            # Postconditions
            self._evaluate_contract_conditions(state, 'postconditions', step)

            # Notify properties
            self._raise_event(MetaEvent('state exited', state=state.name))

        # Execute transition
        if step.transition:
            # Preconditions and invariants
            self._evaluate_contract_conditions(step.transition, 'preconditions', step)
            self._evaluate_contract_conditions(step.transition, 'invariants', step)

            sent_events.extend(self._evaluator.execute_action(step.transition, step.event))

            # Postconditions and invariants
            self._evaluate_contract_conditions(step.transition, 'postconditions', step)
            self._evaluate_contract_conditions(step.transition, 'invariants', step)

            # Update idle time
            self._idle_time[step.transition.source] = self.time

            # Notify properties
            self._raise_event(MetaEvent(
                'transition processed',
                source=step.transition.source,
                target=step.transition.target,
                event=step.event
            ))

        # Enter states
        for state in entered_states:
            # Preconditions
            self._evaluate_contract_conditions(state, 'preconditions', step)

            # Execute entry action
            sent_events.extend(self._evaluator.execute_on_entry(state))

            # Update configuration
            self._configuration.add(state.name)
            self._entry_time[state.name] = self.time
            self._idle_time[state.name] = self.time

            # Notify properties
            self._raise_event(MetaEvent('state entered', state=state.name))

        # Send events
        for event in cast(Union[InternalEvent, MetaEvent], sent_events):
            self._raise_event(event)
            self._sent_events.append(event)

        return MicroStep(event=step.event, transition=step.transition,
                         entered_states=step.entered_states, exited_states=step.exited_states,
                         sent_events=sent_events)

    def _stabilize(self) -> List[MicroStep]:
        """
        Compute, apply and return stabilization steps.

        :return: A list of applied  *MicroStep* instances,
        """
        # Stabilization
        steps = []
        step = self._create_stabilization_step(self._configuration)
        while step is not None:
            steps.append(self._apply_step(step))
            step = self._create_stabilization_step(self._configuration)
        return steps

    def _evaluate_contract_conditions(self, obj: Union[Transition, StateMixin],
                                      cond_type: str,
                                      step: Optional[Union[MacroStep, MicroStep]] = None) -> None:
        """
        Evaluate the conditions for given object.

        :param obj: object with preconditions, postconditions or invariants
        :param cond_type: either "preconditions", "postconditions" or "invariants"
        :param step: step in which the check occurs.
        :raises ContractError: if a condition fails and *ignore_contract* is False.
        """
        if self._ignore_contract:
            return

        exception_klass = cast(Callable[..., Exception], {'preconditions': PreconditionError,
                                                          'postconditions': PostconditionError,
                                                          'invariants': InvariantError}[cond_type])

        unsatisfied_conditions = getattr(
            self._evaluator, 'evaluate_' + cond_type)(obj, getattr(step, 'event', None))

        for condition in unsatisfied_conditions:
            raise exception_klass(configuration=self.configuration, step=step, obj=obj,
                                  assertion=condition, context=self.context)

    def __repr__(self):
        return '{}({!r})'.format(self.__class__.__name__, self._statechart)



class SismicError(Exception):
    pass


class StatechartError(SismicError):
    """
    Base error for anything that is related to a statechart.
    """
    pass


class CodeEvaluationError(SismicError):
    """
    Base error for anything related to the evaluation of the code contained in a statechart.
    """
    pass


class ExecutionError(SismicError):
    """
    Base error for anything related to the execution of a statechart.
    """
    pass


class ConflictingTransitionsError(ExecutionError):
    """
    When multiple conflicting (parallel) transitions can be processed at the same time.
    """
    pass


class NonDeterminismError(ExecutionError):
    """
    In case of non-determinism.
    """
    pass


class PropertyStatechartError(SismicError):
    """
    Raised when a property statechart reaches a final state.

    :param property_statechart: the property statechart that reaches a final state
    """

    def __init__(self, property_statechart):
        super().__init__()
        self._property = property_statechart

    @property
    def property_statechart(self):
        return self._property

    def __str__(self):
        return '{}\nProperty is not satisfied, {} has reached a final state'.format(
            self.__class__.__name__, self._property)


class ContractError(SismicError):
    """
    Base exception for situations in which a contract is not satisfied.
    All the parameters are optional, and are exposed to ease debug.

    :param configuration: list of active states
    :param step: a *MicroStep* or *MacroStep* instance.
    :param obj: the object that is concerned by the assertion
    :param assertion: the assertion that failed
    :param context: the context in which the condition failed
    """

    __slots__ = ['_configuration', '_step', '_obj', '_assertion', '_context']

    def __init__(self, configuration=None, step=None, obj=None, assertion=None, context=None):
        super().__init__(self)
        self._configuration = configuration
        self._step = step
        self._obj = obj
        self._assertion = assertion
        self._context = context

    @property
    def configuration(self):
        return self._configuration

    @property
    def step(self):
        return self._step

    @property
    def obj(self):
        return self._obj

    @property
    def condition(self):
        return self._assertion

    @property
    def context(self):
        return self._context

    def __str__(self):  # pragma: no cover
        message = ['{}'.format(self.__class__.__name__)]
        if self._obj:
            message.append('Object: {}'.format(self._obj))
        if self._assertion:
            message.append('Assertion: {}'.format(self._assertion))
        if self._configuration:
            message.append('Configuration: {}'.format(self._configuration))
        if self._step:
            message.append('Step: {}'.format(self._step))
        if self._context:
            message.append('Context:')
            for key, value in sorted(self._context.items(), key=lambda t: t[0]):
                message.append(' - {key} = {value}'.format(key=key, value=value))

        return '\n'.join(message)


class PreconditionError(ContractError):
    """
    A precondition is not satisfied.
    """
    pass


class PostconditionError(ContractError):
    """
    A postcondition is not satisfied.
    """
    pass


class InvariantError(ContractError):
    """
    An invariant is not satisfied.
    """
    pass



import pytest

from time import sleep

from sismic.runner import AsyncRunner
from sismic.interpreter import Interpreter


class TestAsyncRunner:
    INTERVAL = 0.02

    @pytest.fixture()
    def interpreter(self, simple_statechart):
        return Interpreter(simple_statechart)

    @pytest.fixture()
    def runner(self, interpreter):
        r = AsyncRunner(interpreter, interval=0)
        yield r
        r.stop()

    @pytest.fixture()
    def mocked_runner(self, interpreter, mocker):
        class MockedRunner(AsyncRunner):
            before_run = mocker.MagicMock()
            before_execute = mocker.MagicMock()
            after_execute = mocker.MagicMock()
            after_run = mocker.MagicMock()

        r = MockedRunner(interpreter, interval=0)
        yield r
        r.stop()

    def test_not_yet_started(self, runner):
        assert runner.interpreter.configuration == []

        runner.interpreter.queue('goto s2')
        sleep(self.INTERVAL)
        assert runner.interpreter.configuration == []

        runner.start()
        sleep(self.INTERVAL)
        assert runner.interpreter.configuration == ['root', 's3']

    def test_start(self, runner):
        runner.start()
        sleep(self.INTERVAL)
        assert runner.interpreter.configuration == ['root', 's1']

        runner.interpreter.queue('goto s2')
        sleep(self.INTERVAL)
        assert runner.interpreter.configuration == ['root', 's3']

    def test_restart_stopped(self, runner):
        runner.start()
        runner.stop()

        with pytest.raises(RuntimeError, match='Cannot restart'):
            runner.start()

        assert not runner.running

    def test_start_again(self, runner):
        runner.start()

        with pytest.raises(RuntimeError, match='already started'):
            runner.start()

        assert runner.running
        runner.stop()
        assert not runner.running

    def test_hooks(self, mocked_runner):

        assert len(mocked_runner.before_run.call_args_list) == 0
        assert len(mocked_runner.before_execute.call_args_list) == 0
        assert len(mocked_runner.after_execute.call_args_list) == 0

        mocked_runner.start()
        sleep(self.INTERVAL)
        assert len(mocked_runner.before_run.call_args_list) == 1

        sleep(self.INTERVAL)

        assert len(mocked_runner.before_execute.call_args_list) > 0
        assert len(mocked_runner.after_execute.call_args_list) > 0

        assert len(mocked_runner.after_run.call_args_list) == 0
        mocked_runner.stop()
        sleep(self.INTERVAL)
        assert len(mocked_runner.after_run.call_args_list) == 1


    def test_final(self, runner):
        runner.start()
        runner.interpreter.queue('goto s2')
        runner.interpreter.queue('goto final')
        sleep(self.INTERVAL)
        sleep(self.INTERVAL)
        assert runner.interpreter.final
        assert not runner.running
        sleep(self.INTERVAL)  # Wait for the thread to finish
        assert not runner._thread.is_alive()

    def test_pause(self, runner):
        runner.start()
        assert not runner.paused

        sleep(self.INTERVAL)
        runner.pause()
        assert runner.paused
        assert runner.running
        assert runner.interpreter.configuration == ['root', 's1']

        runner.interpreter.queue('goto s2')
        sleep(self.INTERVAL)
        assert runner.interpreter.configuration == ['root', 's1']

        runner.unpause()
        assert not runner.paused
        assert runner.running
        sleep(self.INTERVAL)
        assert runner.interpreter.configuration == ['root', 's3']


    def test_state(self, runner):
        assert not runner.running
        assert not runner.paused
        runner.start()
        assert runner.running
        assert not runner.paused
        runner.pause()
        assert runner.running
        assert runner.paused
        runner.unpause()
        assert runner.running
        assert not runner.paused
        runner.stop()
        assert not runner.running
        assert not runner.paused

    def test_join_stopped(self, runner):
        runner.start()
        runner.stop()
        runner.wait()



import pytest

from time import sleep
from sismic.clock import SimulatedClock, UtcClock, SynchronizedClock


class TestSimulatedClock:
    @pytest.fixture()
    def clock(self):
        return SimulatedClock()

    def test_initial_value(self, clock):
        assert clock.time == 0

    def test_manual_increment(self, clock):
        clock.time += 1
        assert clock.time == 1

    def test_monotonicity(self, clock):
        clock.time = 10
        with pytest.raises(ValueError):
            clock.time = 0

    def test_automatic_increment(self, clock):
        clock.start()
        sleep(0.1)
        assert clock.time >= 0.1

        clock.stop()
        value = clock.time
        sleep(0.1)
        assert clock.time == value

    def test_increment(self, clock):
        clock.start()
        sleep(0.1)
        assert clock.time >= 0.1
        clock.time = 10
        assert 10 <= clock.time < 10.1
        clock.stop()
        clock.time = 20
        assert clock.time == 20

    def test_speed_with_manual(self, clock):
        clock.speed = 2
        clock.time += 10
        assert clock.time == 10

    def test_speed_with_automatic(self, clock):
        clock.speed = 2
        clock.start()
        sleep(0.1)
        assert clock.time >= 0.2

        clock.stop()
        clock.time = 10
        clock.speed = 0.1
        
        clock.start()
        sleep(0.1)
        clock.stop()
        
        assert 10 < clock.time < 10.1

    def test_start_stop(self, clock):
        clock.start()
        sleep(0.1)
        clock.stop()
        sleep(0.1)
        clock.start()
        sleep(0.1)
        clock.stop()

        assert 0.2 <= clock.time < 0.3


class TestUtcClock:
    @pytest.fixture()
    def clock(self):
        return UtcClock()

    def test_increase(self, clock):
        current_time = clock.time
        sleep(0.1)
        assert clock.time > current_time


class TestSynchronizedClock():
    @pytest.fixture()
    def interpreter(self, mocker):
        interpreter = mocker.MagicMock()
        interpreter.time = 0
        return interpreter
    
    @pytest.fixture()
    def clock(self, interpreter):
        return SynchronizedClock(interpreter)

    def test_init(self, clock):
        assert clock.time == 0

    def test_sync(self, clock, interpreter):
        interpreter.time = 3
        assert clock.time == 3

    def test_no_sync_with_clock(self, clock, interpreter):
        interpreter.clock.time = 3
        assert clock.time == 0
        


from os import path
from codecs import open
from setuptools import setup, find_packages

import sismic

# Get the long description from the README file
with open(path.join(path.abspath(path.dirname(__file__)), 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name=sismic.__package__,
    version=sismic.__version__,
    license=sismic.__licence__,

    author=sismic.__author__,
    url=sismic.__url__,

    description=sismic.__description__,
    long_description=long_description,

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        'Development Status :: 5 - Production/Stable',

        'Environment :: Console',

        'Intended Audience :: Developers',
        'Intended Audience :: Education',
        'Intended Audience :: Information Technology',
        'Intended Audience :: Science/Research',

        'Topic :: Software Development :: Interpreters',
        'Topic :: Software Development :: Testing',
        'Topic :: Scientific/Engineering',

        'License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)',

        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
        'Programming Language :: Python :: 3.13',
    ],
    keywords='statechart state machine interpreter model uml scxml harel',

    packages=find_packages(exclude=['docs', 'tests']),
    python_requires='>=3.9',
    install_requires=[
        'ruamel.yaml>=0.18.2',
        'schema>=0.5.0',
        'behave>=1.2.6',
    ],

    entry_points={
        'console_scripts': [
            'sismic-bdd=sismic.bdd.__main__:cli',
            'sismic-plantuml=sismic.io.plantuml:cli',
        ],
    },

    include_package_data=True,
    zip_safe=True,
)
